package cz.aprar.bonitoo.inference;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import ml.dmlc.xgboost4j.java.Booster;
import ml.dmlc.xgboost4j.java.DMatrix;
import ml.dmlc.xgboost4j.java.XGBoost;
import ml.dmlc.xgboost4j.java.XGBoostError;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Map;

public class CacheInference {
    private static final TTL[] TTL_VALUES = TTL.values();
    private static final int NUM_FEATURES = 20;

    private final ObjectMapper mapper = new ObjectMapper();
    private final Booster booster;
    private final Map<String, Map<String, Integer>> labels;

    public CacheInference(final Path modelFile, final Path labelFile) throws IOException, XGBoostError {
        try (BufferedInputStream modelIS = new BufferedInputStream(Files.newInputStream(modelFile));
             BufferedInputStream labelIS = new BufferedInputStream(Files.newInputStream(labelFile))
        ) {
            booster = loadModel(modelIS);
            labels = loadLabels(labelIS);
        }
    }

    public CacheInference(InputStream modelIS, InputStream labelIS) throws IOException, XGBoostError {
        booster = loadModel(modelIS);
        labels = loadLabels(labelIS);
    }

    public TTL[] cacheTTL(final FlightData[] data) throws XGBoostError {
        return cacheTTL(data, ZonedDateTime.now());
    }

    /**
     * Method for backtesting
     */
    public TTL[] cacheTTL(final FlightData[] data, final ZonedDateTime now) throws XGBoostError {
        final float[][] predicts = cacheTTLProbability(data, now);

        if (predicts.length == 0) {
            return new TTL[] {};
        }

        final TTL[] ttl = new TTL[predicts.length];
        for (int i = 0; i < predicts.length; i++) {
            int idx = 0;
            float max = Float.MIN_VALUE;

            for (int j = 0; j < predicts[i].length; j++) {
                if (predicts[i][j] > max) {
                    max = predicts[i][j];
                    idx = j;
                }
            }

            ttl[i] = TTL_VALUES[idx];
        }

        return ttl;
    }

    public float[][] cacheTTLProbability(final FlightData[] data) throws XGBoostError {
        return cacheTTLProbability(data, ZonedDateTime.now());
    }

    /**
     * Method for backtesting
     */
    public float[][] cacheTTLProbability(final FlightData[] data, final ZonedDateTime now) throws XGBoostError {
        final DMatrix matrix = createMatrix(data, now);
        return booster.predict(matrix);
    }

    private Booster loadModel(InputStream model) throws XGBoostError, IOException {
        return XGBoost.loadModel(model);
    }

    private DMatrix createMatrix(final FlightData[] data, final ZonedDateTime now) throws XGBoostError {
        final float[] arr = new float[data.length * NUM_FEATURES];
        for (int i = 0; i < data.length; i++) {
            final FlightData current = data[i];
            final int start = NUM_FEATURES * i;
            arr[start] = (float) current.getInputPrice();
            arr[start + 1] = labels.get("type").getOrDefault(current.getType(), 0);
            arr[start + 2] = labels.get("flight.inboundSegments.departure").getOrDefault(current.getInboundDeparture(), 0);
            arr[start + 3] = labels.get("flight.inboundSegments.arrival").getOrDefault(current.getInboundArrival(), 0);
            arr[start + 4] = labels.get("flight.inboundSegments.origin.airportCode").getOrDefault(current.getInboundOrigin(), 0);
            arr[start + 5] = labels.get("flight.inboundSegments.destination.airportCode").getOrDefault(current.getInboundDestination(), 0);
            arr[start + 6] = labels.get("flight.inboundSegments.airline.code").getOrDefault(current.getInboundAirlines(), 0);
            arr[start + 7] = labels.get("flight.inboundMCX.code").getOrDefault(current.getInboundMCXAirlines(), 0);
            arr[start + 8] = labels.get("flight.outboundSegments.departure").getOrDefault(current.getOutboundDeparture(), 0);
            arr[start + 9] = labels.get("flight.outboundSegments.arrival").getOrDefault(current.getOutboundArrival(), 0);
            arr[start + 10] = labels.get("flight.outboundSegments.origin.airportCode").getOrDefault(current.getOutboundOrigin(), 0);
            arr[start + 11] = labels.get("flight.outboundSegments.destination.airportCode").getOrDefault(current.getOutboundDestination(), 0);
            arr[start + 12] = labels.get("flight.outboundSegments.airline.code").getOrDefault(current.getOutboundAirlines(), 0);
            arr[start + 13] = labels.get("flight.outboundMCX.code").getOrDefault(current.getOutboundMCXAirlines(), 0);
            arr[start + 14] = current.getNights();
            arr[start + 15] = computePrebooking(current.getOutDepartureDate(), now);
            arr[start + 16] = now.getDayOfMonth();
            arr[start + 17] = now.getDayOfWeek().getValue() - 1;
            arr[start + 18] = current.getOutDepartureDate().getDayOfMonth();
            arr[start + 19] = current.getOutDepartureDate().getDayOfWeek().getValue() - 1;
        }

        return new DMatrix(arr, data.length, NUM_FEATURES);
    }

    private float computePrebooking(final ZonedDateTime outdeparture, final ZonedDateTime now) {
        return ChronoUnit.DAYS.between(now, outdeparture);
    }

    private Map<String, Map<String, Integer>> loadLabels(InputStream labels) throws IOException {
        final TypeReference<Map<String, Map<String, Integer>>> typeRef = new TypeReference<Map<String, Map<String, Integer>>>() {
        };
        return mapper.readValue(labels, typeRef);
    }
}
